"""Embedded HTTP Server.

This module builds on BaseHTTPServer by implementing the standard GET
and HEAD requests in a fairly straightforward manner, dispatching
them to pre-registered handlers for paths.  This allows the creation of
a virtual URL-space for the embedded webserver.

TO-DO: POST.

"""

__version__ = "0.1"
__author__ = "Mark Moraes <http://www.cs.toronto.edu/~moraes>"
__copyright__ = "Copyright 2003, Mark Moraes"
__license__ = "Python"

__all__ = ["EmbeddedHTTPRequestHandler"]

import sys, traceback, posixpath
import BaseHTTPServer
import urllib

debug = 0

# MODIFICATION FOR PEARS
try:
    from pearsconstants import *
    debug = DEBUGMODE
except:
    pass
    

objmap = {}

def register(path, handler, *args, **kwargs):
    """Register a handler for the specified path."""
    objmap[path] = (handler, args[:], kwargs.copy())

def handle(path, urlargs):
    """Call a handler for a specific path with the dict urlargs
    A handler must return an object which can be stringified for printing
    with str() and whose stringified length will be returned when len()
    is called on it. (i.e. must properly implement __str__ and __len__).
    If it provides an http_type() method, that must return the MIME content
    type (e.g. 'text/html' if an HTML page is being returned.)  If such
    a method is not provided, 'text/plain' is assumed.
    handler can also be a string, in which case it is treated as a format
    specification and interpolated with the urlargs.
    _SCRIPT_NAME_ is the full path of the object while
    _SCRIPT_BASENAME_ is the final component of the path.
    """
    # shorten path by one component on each search till we find one.
    p = path
    while p not in objmap:
        pnew = posixpath.dirname(p)
        if pnew == '' or pnew == p:
            break
        p = pnew

    if p in objmap:
        handler, args, kwargs = objmap[p]
        kw = kwargs.copy()
        kw.update(urlargs)
        kw['_SCRIPT_NAME_'] = path
        kw['_SCRIPT_PATHNAME_'] = p
        kw['_SCRIPT_RESTNAME_'] = path[len(p)+1:]
        kw['_SCRIPT_BASENAME_'] = posixpath.basename(path)
        try:
            if callable(handler):
                return handler(*args, **kw)
            else:
                return handler % kw
        except:
            etype, value, tb = sys.exc_info()
            return "Error with handler for "+`path`+': '+`traceback.format_exception_only(etype, value)`
    else:
        return None

# http://127.0.0.1:5335/system/pages/subscriptions?url=http://www.rolandTanglao.com/rss.xml
class EmbeddedHTTPRequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):

    """Simple HTTP request handler with GET and HEAD commands.

    This subscribes to a feed described in the url parameter.

    """

    server_version = "EmbeddedHTTP/" + __version__

    def do_GET(self):
        """Serve a GET request."""
        if debug:
            print >> sys.stderr, "GET", `self.path`
        f = self.send_head()
        if f:
            self.wfile.write(str(f))

    def do_HEAD(self):
        """Serve a HEAD request."""
        if debug:
            print >> sys.stderr, "HEAD", `self.path`
        self.send_head()

    def send_head(self):
        """Common code for GET and HEAD commands.

        This sends the response code and MIME headers.

        Return value is an object (whose str() form has to be copied
        to the outputfile by the caller unless the command was HEAD)
        or None, in which case the caller has nothing further to do.

        """
        obj = self.translate_path(self.path)
        if obj:
            if hasattr(obj, 'http_type'):
                ctype = obj.http_type()
            else:
                ctype = 'text/plain'
            self.send_response(200)
            self.send_header("Content-type", ctype)
            self.send_header("Content-Length", str(len(obj)))
            self.end_headers()
        else:
            self.send_error(404, "File not found")
        return obj

    def translate_path(self, path):
        """Translate a /-separated PATH to an object.
        
        Must return an object that supports str() and len().  Optionally
        http_type() if it wishes a type other than "text/plain"
        """
        path = urllib.unquote(path)
        # convert to an object via lookup in some sort of registry.
        if debug:
            print >> sys.stderr, "path", `path`
        i = path.find("?")
        if i >= 0:
            urlargs = {}
            for arg in path[i+1:].split('&'):
                j = arg.find('=')
                if j >= 0:
                    urlargs[arg[:j]] = arg[j+1:]
                else:
                    urlargs[arg] = None
            path = path[:i]
        else:
            urlargs = None
        return handle(path, urlargs)


def test(HandlerClass = EmbeddedHTTPRequestHandler,
         ServerClass = BaseHTTPServer.HTTPServer):
    BaseHTTPServer.test(HandlerClass, ServerClass)

if __name__ == '__main__':
    def t1(*args, **kwargs):
        return str(['t1', args, kwargs])
    def t2(*args, **kwargs):
        return str(['t2', args, kwargs])
    print "registering /foo/bar"
    register('/foo/bar', t1)
    print "registering /foo"
    register('/foo', t2, 'defaultarg1', defaultarg2=2)
    print "registering /system/pages/subscriptions"
    register('/system/pages/subscriptions', "subscribe %(url)s")
    test()
