#! /usr/bin/env python
"""This module implements PersistentDict, a persistent dictionary whose
   item values are loaded on demand from and saved automatically
   to pickled files in a filesystem directory.  Essentially, this acts
   as a write-through cache.
"""

# Issues/To-Do:
# Callback after load or before unload of any dict item.
# Wish we could use weakref rather than _maxcached, but weakref
# does not seem to work on int, dict, etc.
# How about deepcopy: need a new dictdir
# Note: if x is a PersistentDict, and x[key1] = x[key2] = foo,
# two copies of foo will be pickled and stored in the files for key1 and key2
# and subsequent instances of this PersistentDict will not be the
# same object.  For a fully general solution, we would need swizzling?

import cPickle as pickle
import md5 as hash
import os
import UserDict

__version__ = "0.1"
__author__ = "Mark Moraes <http://www.cs.toronto.edu/~moraes>"
__copyright__ = "Copyright 2003, Mark Moraes"
__license__ = "Python"

debug = 0
if debug:
    import sys

NOTLOADED = '__NotLoaded__'

def safedump(dir, fname, val):
    """Safe save: writes to a temp file first, then renames to the 
       real name.  Safe because it avoids trashing data if we fail
       the complete write+close e.g. run out of disk space
       WARNING: if dictdir is writable by other users, it is vulnerable
       to race condition attacks.
       Note: leaves a temporary file lying around if the pickle write fails.
    """
    assert val != NOTLOADED
    tf = os.path.join(dir, '_'+fname)       # temporary new file
    if debug:
        print >> sys.stderr, 'PersistentDict dumping to', `tf`
    pickle.dump(val, open(tf,'w'))
    xf = os.path.join(dir, fname)
    bf = ''
    if os.path.exists(xf):
        bf = os.path.join(dir, '__'+fname)  # backup file
        if os.path.exists(bf):
            os.remove(bf)                   # remove previous backup file
        os.rename(xf, bf)
        if debug:
            print >> sys.stderr, 'PersistentDict removing old', `xf`
    if debug:
        print >> sys.stderr, 'PersistentDict renaming', `tf`, 'to', `xf`
    os.rename(tf, xf)
    if bf:
        os.remove(bf)                       # no need for backup file now

# To-Do: ensure that multiple dicts created on the same dictdir
# actually become the same dict?  Or lock the dictdir somehow.

IDXNAME = 'INDEX'
class PersistentDict(UserDict.DictMixin):
    def __init__(self, dictdir,maxcached = 1,newitemfunc=None):
        # dictdir is a directory in which each file contains the value
        # of a dict element, with the filename being the MD5 hash of the key.
        self._dictdir = dictdir
        self._maxcached = maxcached
        self._newitemfunc = newitemfunc
        self._lru = []
        self._keys = []
        self._cache = {}
        if os.path.exists(self._dictdir):
            idx = os.path.join(self._dictdir, IDXNAME)
            if os.path.exists(idx):
                self._keys = pickle.load(open(idx, 'r'))
                self._cache = dict([(hash.new(i).hexdigest(),NOTLOADED)
                                    for i in self._keys])
                if debug:
                    print >> sys.stderr, 'PersistentDict loaded', `self._cache.items()`
        else:
            os.mkdir(self._dictdir)
            if debug:
                print >> sys.stderr, 'PersistentDict mkdir', `self._dictdir`
    def __getinitargs__(self):
        """If PersistentDict is being unpickled, we want __init__()
           called with the dictdir name
        """
        return (self._dictdir,self._maxcached)
    def __getstate__(self):
        """Only need these fields to be pickled, rest is persistent anyway"""
        return (self._dictdir,self._maxcached)
    def __setstate__(self, state):
        """__init__ will handle unpickling"""
        pass
    def _flushitem(self, h):
        """Drops item from cache"""
        if self._cache[h] != NOTLOADED:
            safedump(self._dictdir, h, self._cache[h])
            self._cache[h] = NOTLOADED
        try:
            self._lru.remove(h)
        except ValueError:
            pass
        if debug:
            print >> sys.stderr, 'PersistentDict _flushitem', `h`
    def flushitem(self, key):
        """Remove the value for the specified key from memory cache"""
        self._flushitem(hash.new(key).hexdigest())
    def flushexcept(self, keeplist):
        """Removes values from memory for all keys except those in keeplist"""
        keephashes = [hash.new(key).hexdigest() for key in keeplist]
        for h in self._cache.keys():
            if h not in keephashes:
                self._flushitem(h)
            elif debug:
                print >> sys.stderr, "PersistentDict keeping", `h`
    def sync(self):
        """Saves all items to disk.  No need to save IDXNAME because it
           is always synced whenever changed.
        """
        for h in self._cache.keys():
            if self._cache[h] != NOTLOADED:
                safedump(self._dictdir, h, self._cache[h])
    def clear(self, removedir = False):
        """Removes all items.  If removedir is set, then also
           removes the underlying directory (after which this object
           cannot be used)
        """
        # first empty IDXNAME so that even if unlinks/rmdir fail or are
        # interrupted, the persistent dict is clear thereafter.
        self._keys = []
        self._lru = []
        safedump(self._dictdir, IDXNAME, [])
        for h in self._cache.keys():
            os.remove(os.path.join(self._dictdir, h))
        if removedir:
            os.remove(os.path.join(self._dictdir, IDXNAME))
            os.rmdir(self._dictdir)
            self._cache = None
            self._keys = None
            # NOTE: CANNOT use this object after calling clear(removedir=True)
        else:
            self._cache = {}
    def size(self):
        """Returns estimate of total number of bytes used by dict on disk"""
        n = 0
        for f in os.listdir(self._dictdir):
            n += os.path.getsize(os.path.join(self._dictdir, f))
        return n
    def _mru(self, h):
        """Indicate h is most-recently-used by putting it at the 
           end of the _lru list
        """
        if h in self._lru:
            self._lru.remove(h)
        self._lru.append(h)
    def __contains__(self, key):
        # DictMixin calls has_key, which loads the item by default!
        return key in self._keys
    has_key = __contains__
    def __getitem__(self, key):
        h = hash.new(key).hexdigest()
        if h not in self._cache:
            raise KeyError, key
        elif self._cache[h] == NOTLOADED:
            if len(self._lru) >= self._maxcached:
                if debug:
                    print >> sys.stderr, 'PersistentDict uncaching', `self._lru[0]`
                self._flushitem(self._lru.pop(0))
            fname = os.path.join(self._dictdir, h)
            if os.path.exists(fname):
                val = pickle.load(open(fname,'r'))
            else:
                # XXX check for backup file?
                if self._newitemfunc:
                    val = self._newitemfunc(key)
                else:
                    val = None
            self._cache[h] = val
            if debug:
                print >> sys.stderr, 'PersistentDict loaded', `h`, `key`
        self._mru(h)
        return self._cache[h]
    def __setitem__(self, key, value):
        h = hash.new(key).hexdigest()
        safedump(self._dictdir, h, value)
        if h not in self._cache:
            self._keys.append(key)
            safedump(self._dictdir, IDXNAME, self._keys)
        self._mru(h)
        self._cache[h] = value
    def __delitem__(self, key):
        h = hash.new(key).hexdigest()
        if h not in self._cache:
            raise KeyError, key
        else:
            del self._cache[h]
            self._keys.remove(key)
            try:
                self._lru.remove(h)
            except ValueError:
                pass
            safedump(self._dictdir, IDXNAME, self._keys)
            os.remove(os.path.join(self._dictdir, h))
            if debug:
                print >> sys.stderr, 'PersistentDict deleted', `h`, `key`
    def keys(self):
        if debug:
            print >> sys.stderr, 'PersistentDict keys', `self._cache.keys()`, `self._keys`
        return self._keys

if __name__ == '__main__':
    tdirx = 'test_dir_x_will_be_removed'
    tdiry = 'test_dir_y_will_be_removed'
    x = PersistentDict(tdirx)
    print 'init', x.keys()
    x['hello'] = 1
    x['world'] = 2
    x.flushexcept([])
    print 'after flush', x.keys()
    assert 'hello' in x
    assert x.has_key('hello')
    assert x['hello'] == 1
    x.flushitem('hello')
    del x
    x = PersistentDict(tdirx)
    for i in x.keys():
        print `i`, `x[i]`
        x[i] += 10
        print ' ->', `x[i]`
    y = PersistentDict(tdiry)
    y['junk'] = x
    assert y['junk']['hello'] == 11
    assert y['junk']['world'] == 12
    y.sync()
    del x
    print `tdirx`, `os.listdir(tdirx)`
    print `tdiry`, `os.listdir(tdiry)`, y.size()
    x = PersistentDict(tdirx)
    print x.items()
    y = PersistentDict(tdiry)
    print y.items()
    x.clear()
    assert len(x.keys()) == 0
    x['test'] = 3
    x.clear(True)
    y.clear(True)
    try:
        x['will fail'] = 1
    except IOError:
        pass
        
