#
#  Name:        ServerMixin
#  This was modified from Greg Landrum's modified version
#  http://www.mindspring.com/~greglandrum/python/rpcMixin.py
#  of rpcMixin.py distributed with wxPython
#
#  TO-DO: demonstrate that this can be used with xmlrpc as well as
#  http (or other socket servers?)
#  
#----------------------------------------------------------------------

__version__ = "0.1"
__author__ = """
    greg Landrum <Landrum@RationalDiscovery.com>
    Mark Moraes <http://www.cs.toronto.edu/~moraes>
    """
__copyright__ = """
    Copyright:   (c) 2000, 2001 by Greg Landrum and Rational Discovery LLC
    Copyright 2003, Mark Moraes
    """
__license__ = "wxWindows"

"""ServerMixin provides server functionality for wxPython applications via a
   mixin class

**Some Notes:**

  1)  The server runs in a separate thread from the main GUI
      application, communication between the two threads using a custom
      event (see the Threads demo in the wxPython docs for more info).
      Requires:    Python with threading enabled.
  2)  See the bottom of this file for an example of using the class.
  
"""


from wxPython.wx import *
# import xmlrpcserver,xmlrpclib
import threading
import SocketServer
import string
import new
import sys
try:
    from timedoutsocket import Timeout
except:
    # seems suitable
    Timeout = EOFError

rpcPENDING = 0
rpcDONE = 1
rpcEXCEPT = 2

class RPCRequest:
  """A wrapper to use for handling requests and their responses"""
  status = rpcPENDING
  result = None

# here's the ID for external events
wxEVT_EXTERNAL_EVENT = 25015
class ExternalEvent(wxPyEvent):
  """The custom event class used to pass xmlrpc calls from
     the server thread into the GUI thread

  """
  def __init__(self,method,*args,**kwargs):
    wxPyEvent.__init__(self)
    self.SetEventType(wxEVT_EXTERNAL_EVENT)
    self.method = method
    self.args = args
    self.kwargs = kwargs
    self.rpcStatus = RPCRequest()
    self.rpcStatusLock = threading.Lock()
    self.rpcCondVar = threading.Condition()

  def Destroy(self):
    self.method=None
    self.args=None
    self.rpcStatus = None
    self.rpcStatusLock = None
    self.rpcondVar = None
    
def EVT_EXTERNAL_EVENT(win,func):
  win.Connect(-1,-1,wxEVT_EXTERNAL_EVENT,func)

# Send an event to the GUI with some defined params. (args, kwargs)
# This is essentially an intra-thread RPC.  The request handlers for the
# server objects should call this method whenever they need the GUI's
# attention.
# e.g. if frame is a wxPython object with the ServerMixin, then
# the server handler should call notifyGUI(frame, method, **kwargs)
# whenever the frame needs to be aware of any action, which will
# result in frame.method(**kwargs) being called.
#
def notifyGUI(win, method, *args, **kwargs):
    """Tell the GUI we have an event.
    """
    # construct the event
    evt = ExternalEvent(method,*args,**kwargs)

    # update the status variable
    evt.rpcStatusLock.acquire()
    evt.rpcStatus.status = rpcPENDING
    evt.rpcStatusLock.release()

    evt.rpcCondVar.acquire()
    # dispatch the event to the GUI
    # print 'posting event to', `win`
    wxPostEvent(win,evt)

    # wait for the GUI to finish
    while evt.rpcStatus.status == rpcPENDING:
      evt.rpcCondVar.wait()
    evt.rpcCondVar.release()
    evt.rpcStatusLock.acquire()
    if evt.rpcStatus.status == rpcEXCEPT:
      # The GUI threw an exception, release the status lock
      #  and re-raise the exception
      evt.rpcStatusLock.release()
      raise evt.rpcStatus.result[0],evt.rpcStatus.result[1]
    else:
      # everything went through without problems
      s = evt.rpcStatus.result

      evt.rpcStatusLock.release()
      evt.Destroy()
      return s

# this global Event is used to let the server thread
#  know when it should quit
stopEvent = threading.Event()
stopEvent.clear()

class _ServerThread(threading.Thread):
  """ this is the Thread class which actually runs the server

  """
  def __init__(self,server,verbose=0):
    self._Serv = server
    threading.Thread.__init__(self,verbose=verbose)

  def stop(self):
    stopEvent.set()

  def shouldStop(self):
    return stopEvent.isSet()

  def run(self):
    while not self.shouldStop():
      # if handle_request is blocking on I/O, how do we get out of
      # that to check if we need to stop?  If we knew the protocol,
      # then we could do an internal connect, ick!
      try:
        self._Serv.handle_request()
      except Timeout:
        time.sleep(0.1)
      except:
        pass
    self._Serv = None
    
class ServerMixin:
  """A mixin class to provide server functionality to wxPython
     frames/windows

     If you want to customize this, probably the best idea is to
     override the OnExternal method, which is what's invoked when an
     RPC is handled.

  """
  def __init__(self,server,serverfinish,name=None,verbose=0):
    """Constructor
    
      **Arguments**
        - verbose: (optional) if set, the server thread will be launched
          in verbose mode
    """
    print self
    self.verbose=verbose
    self.serverfinish=serverfinish
    EVT_EXTERNAL_EVENT(self,self.OnExternal)
    if hasattr(self,'OnClose'):
      self._origOnClose = self.OnClose
      self.Disconnect(-1,-1,wxEVT_CLOSE_WINDOW)
    else:
      self._origOnClose = None
    self.OnClose = self.RPCOnClose
    EVT_CLOSE(self,self.RPCOnClose)

    self.servThread = _ServerThread(server,verbose=self.verbose)
    if not name:
        name = 'Handler%d' % server.fileno()
    self.servThread.setName(name)
    self.servThread.start()
    
  def RPCOnClose(self,event):
    """ callback for when the application is closed

       be sure to shutdown the server and the server thread before
       leaving

    """
    # by setting the global stopEvent we inform the server thread
    # that it's time to shut down.
    stopEvent.set()
    if event is not None:
      # if we came in here from a user event (as opposed to an RPC event),
      #  then we'll need to kick the server one last time in order
      #  to get that thread to terminate.  do so now
      try:
        self.serverfinish()
      except:
        pass
    if self._origOnClose is not None:
      self._origOnClose(event)

  def RPCQuit(self):
    """ shuts down everything, including the rpc server

    """
    self.RPCOnClose(None)
  def OnExternal(self,event):
    """ this is the callback used to handle RPCs

      **Arguments**

        - event: an _ExternalEvent_ sent by the rpc server
      
      Exceptions are caught and returned in the rpcStatus
      structure.  This allows the xmlrpc server to report the
      exception to the client without mucking up any of the delicate
      thread stuff.

    """
    event.rpcStatusLock.acquire()
    try:
      methsplit = string.split(event.method,'.')
      meth = self
      for piece in methsplit:
        meth = getattr(meth,piece)
    except AttributeError,msg:
      event.rpcStatus.result = 'No Such Method',msg
      event.rpcStatus.status = rpcEXCEPT
    else:
      try:
        res = apply(meth,event.args,event.kwargs)
      except:
        import traceback
        if self.verbose: traceback.print_exc()
        event.rpcStatus.result = sys.exc_info()[:2]
        event.rpcStatus.status = rpcEXCEPT
      else:
        if res is None:
          # returning None across the xmlrpc interface is problematic
          event.rpcStatus.result = []
        else:
          event.rpcStatus.result = res
        event.rpcStatus.status = rpcDONE
          
    event.rpcStatusLock.release()

    # broadcast (using the condition var) that we're done with the event
    event.rpcCondVar.acquire()
    event.rpcCondVar.notify()
    event.rpcCondVar.release()


if __name__ == '__main__':
  import time
  port = 5335
  import EmbeddedHTTPServer, BaseHTTPServer, httplib
    
  class serverFrame(wxFrame,ServerMixin):
    """A simple wxFrame with the ServerMixin functionality added
    """
    def __init__(self,*args,**kwargs):
      """ server and verbose keyword arguments will be passed along to
          the ServerMixin.
      """
      mixinArgs = {}
      for i in ('server', 'serverfinish', 'verbose'):
        if kwargs.has_key(i):
            mixinArgs[i] = kwargs[i]
            del kwargs[i]
      apply(wxFrame.__init__,(self,)+args,kwargs)
      apply(ServerMixin.__init__,(self,),mixinArgs)
      EVT_CHAR(self,self.OnChar)

    def TestFunc(self,*args,**kwargs):
      """a demo method"""
      s = 'TestFunc:'+`(args, kwargs)`
      print s
      self.SetTitle(s)
      return "ok:"+s

    def OnChar(self,event):
      key = event.GetKeyCode()
      if key == ord('q'):
        self.OnQuit(event)

    def OnQuit(self,event):
      self.OnClose(event)

    def OnClose(self,event):
      self.Destroy()

  class MyApp(wxApp):
    def OnInit(self):
      serv = BaseHTTPServer.HTTPServer(
                    ('localhost',port),
                    EmbeddedHTTPServer.EmbeddedHTTPRequestHandler
               )
      # XXX make dummy final connection to force server thread to exit!
      servfin = lambda x='localhost',y=port: httplib.HTTP().connect(x,y)
      self.frame = serverFrame(NULL, -1, "wxPython RPCDemo", wxDefaultPosition,
                            wxSize(300,300), server=serv,
                            serverfinish=servfin)
      EmbeddedHTTPServer.register('/foo', notifyGUI, self.frame, 'TestFunc')
      self.frame.Show(TRUE)
      return TRUE

  def testcon(port):
    h = httplib.HTTP()
    h.connect('localhost', port)
    h.putrequest('GET', '/foo/bar?message=hello%20world')
    h.endheaders()
    print `h.getreply()`
  
  nT = threading.activeCount()
  app = MyApp(0)
  activePort = port
  t = threading.Thread(target=lambda x=activePort:testcon(x),verbose=0)
  t.start()
  
  app.MainLoop()
  # give the threads time to shut down
  if threading.activeCount() > nT:
    print 'waiting for all threads to terminate'
    while threading.activeCount() > nT:
      time.sleep(0.5)
