#!/usr/bin/env python

"""hamsterengine 1.7 - back end for pears.py 
A patch by Mark Moraes, this engine offers much improved
performance in case of very large cache sizes by using on-demand 
loading.

Copyright (C) 2003, 2004 Project 5

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Contact info
web: http://come.to/project5
mail/msn: project5@wanadoo.nl
icq: 84243714
jabber: project5@rhymbox.com
"""

import os, md5, traceback
from PersistentDict import PersistentDict
from pearsengine import engine as eng
from pearsconstants import *

__version__ = "Hamster PersistentDict (for large cache) - by Mark Moraes"

class engine(eng):
    """Adapted engine with on demand loading."""
    def __init__(self):
        eng.__init__(self)
        self.locked = {}
        self.itemdirname = os.path.join(self.savedir, 'itemdir')
        self.items = PersistentDict(self.itemdirname, maxcached=SIMULTANEOUSUPDATE+1)
        
    def oldItemFile(self, location):
        """Returns old item dir path with MD5 hex representation of location
           as used in v1 of hamsterengine, before use of PersistentDict
        """
        if not hasattr(self, 'olditemdir'):
            self.olditemdir = os.path.join(self.savedir, 'olditemdir')
        return 
        
    def backupFile(self, fname):
        """Rename fname as fname.N.bk"""
        i = 1
        while True: # backup the file
            backupfile = fname + "." + str(i) + ".bk"
            if not os.path.exists(backupfile):
                try:
                    os.rename(fname, backupfile)
                    break
                except:
                    i += 1
            else:
                i += 1
        
    def convertItemsFileToDir(self, itemsfilename):
        """Converts ordinary pears engine file to v2 PersistentDict
           hamsterengine
        """
        items = self.loadSomething(itemsfilename)
        for i in range(0, len(self.feeds)):
            loc = self.feeds[i][LOCATION]
            self.feeds[i][NEWITEMS] = len(items[loc][UNREAD])
            self.feeds[i][TOTITEMS] = len(items[loc][ITEMS])
            self.items[loc] = items[loc]
        self.saveItems()
        self.saveFeeds()
        self.backupFile(itemsfilename)
        
    def convertOldDir(self, itemsdirname):
        """Converts v1 hamsterengine directory to v2 PersistentDict"""
        for i in range(0, len(self.feeds)):
            loc = self.feeds[i][LOCATION]
            fname = os.path.join(itemsdirname, md5.new(loc).hexdigest())
            x = self.loadSomething(fname)
            if x:
                self.feeds[i][NEWITEMS] = len(x[UNREAD])
                self.feeds[i][TOTITEMS] = len(x[ITEMS])
                self.items[loc] = x
                self.backupFile(fname)
        self.saveItems()
        self.saveFeeds()
        self.backupFile(itemsdirname)
        
            
    def loadItems(self):
        """Converts from pearsengine or old v1 hamsterengine format
           and initializes PersistentDict"""
        olditemsfile = os.path.join(self.savedir, 'items.pyrs')
        olditemsdir = os.path.join(self.savedir, 'items')
        if os.path.exists(olditemsfile):
             self.convertItemsFileToDir(olditemsfile)
        elif os.path.exists(olditemsdir):
            self.convertOldDir(olditemsdir)
        self.checkFeeds()
        
    def getTopicCounts(self, feedid):
        loc, id = self.convertid(feedid)
        return self.feeds[id][NEWITEMS], self.feeds[id][TOTITEMS]
        
    def saveItems(self):
        """Saves the items to the items files in the directory."""
        self.items.sync()
        
    def lockFeed(self, feedid):
        """Locks a feed so that it cannot be flushed from memory.
           Use this on a feed whose topics are being displayed"""
        loc, id = self.convertid(feedid)
        if not self.locked.has_key(loc):
            self.locked[loc] = id
            if DEBUGMODE: print 'locked', loc, id
        else:
            print "Feed %s, %d already locked!" % (loc, self.locked[loc])
        return id

    def unlockFeed(self, feedid):
        loc, id = self.convertid(feedid)
        if loc: # loc might be an empty string if feedid has been deleted in the mean time
            if self.locked.has_key(loc):
                if DEBUGMODE: print 'unlocked', loc, id
                del self.locked[loc]
                self.flushFeed(loc)
            else:
                print "Feed %s not locked!" % loc

    def flushFeed(self, loc):
        """Saves a feed and deletes from memory"""
        if not self.locked.has_key(loc):
            self.items.flushitem(loc)
            if DEBUGMODE: print 'flushed', loc
        else:
            if DEBUGMODE: print "Feed %s locked, cannot flush" % loc
            
         
    def updateFeedCounts(self, loc, id, doFlush = False):
        self.feeds[id][NEWITEMS] = len(self.items[loc][UNREAD])
        self.feeds[id][TOTITEMS] = len(self.items[loc][ITEMS])
        if doFlush:
            self.flushFeed(loc)
        
    def isFeedCacheExceeded(self, id):
        """Determines if cache of feed at id (id can be either a string 
           containing the location or an integer indicating the position 
           of the item to be deleted in the list of feeds), is exceeded
           by the current number of items in memory in that feed.
           Returns False if cache not exceeded, True if exceeded."""
        location, index = self.convertid(id)
        return self.feeds[index][CACHESIZE]<self.feeds[index][TOTITEMS]
        
    def getTotalTopicsCount(self):
        """Returns the total number of topics."""
        total = 0
        for feed in self.feeds:
            total += feed[TOTITEMS]
        return total
        
    def getTotalTopicsSize(self):
        """Returns the number of bytes occupied by the saved items."""
        return self.items.size()    
        
    def checkItems(self, id):
        loc, x = self.convertid(id)
        if not self.items.has_key(loc) or not self.items[loc]:
            print "Creating missing items %d %s" % (x, loc)
            self.items[loc] = self.makeEmptyFeed()[1]
            self.feeds[x][ETAG] = None
        
    def getTopics(self, id, newonly=False):
        self.checkItems(id)
        return eng.getTopics(self, id, newonly)
        
    def updateItems(self, id, doforce=True):
        self.checkItems(id)
        return eng.updateItems(self, id, doforce)
        
    def sortFeedsUnread(self):
        def comp(a, b):
            ac = (a[NEWITEMS], a[TOTITEMS])
            bc = (b[NEWITEMS], b[TOTITEMS])
            return -cmp(ac, bc)
        self.feeds.sort(comp)
